# coding=utf-8
from __future__ import unicode_literals
import logging
from collections import OrderedDict
import os
import time
from colorama import Fore, Style
import six
from six import python_2_unicode_compatible
from copy import deepcopy
from .util import pick, utf8print

log = logging.getLogger(__name__)


@python_2_unicode_compatible
class Task(object):
    """
    Task class from which the other task classes inherit.
    """
    def __init__(self):
        self.progress = 0

    def _run(self, marc_record):
        return 0

    def run(self, marc_record, progress=None):
        # log.debug('Run task: %s', self)
        self.progress = progress
        modified = self._run(marc_record)
        return modified


@python_2_unicode_compatible
class SingleSourceConceptTask(Task):

    def __init__(self, source, ignore_extra_subfields):
        super().__init__()
        self.source = deepcopy(source)
        self.ignore_extra_subfields = ignore_extra_subfields

    def match(self, marc_record):
        for field in marc_record.fields:
            if field.tag.startswith('6') and field.match(self.source, self.ignore_extra_subfields):
                return True
        return False


@python_2_unicode_compatible
class MultipleSourceConceptTask(Task):

    def __init__(self, concepts, ignore_extra_subfields):
        super().__init__()
        self.concepts = [deepcopy(concept) for concept in concepts]
        for concept in self.concepts:
            concept.set_a_or_x_to('a')
        self.ignore_extra_subfields = ignore_extra_subfields

    def match_concept(self, marc_record, concept):
        for field in marc_record.fields:
            if field.tag.startswith('6') and field.match(concept, self.ignore_extra_subfields):
                return True
        return False

    def match(self, marc_record):
        for concept in self.concepts:
            if not self.match_concept(marc_record, concept):
                return False
        return True


@python_2_unicode_compatible
class ReplaceTask(SingleSourceConceptTask):
    """
    Replace a subject access or classification number field with another one in
    any given MARC record.
    """

    def __init__(self, source, target, ignore_extra_subfields=False):
        super().__init__(source, ignore_extra_subfields)
        self.target = deepcopy(target)
        self.source.set_a_or_x_to('a')
        self.target.set_a_or_x_to('a')

        """
        Caveat 1:
        If extra subfields are ignored in the matching, we cannot add/update $0,
        since the identifier is connected to the exact subject heading
        (For instance, "$a TermA $b TermB" has a differen $0 than just "$a TermA")
        """
        if self.ignore_extra_subfields:
            if '0' in self.target.sf:
                del self.target.sf['0']

    def __str__(self):
        ign = ' (ignoring any extra subfields)' if self.ignore_extra_subfields else ''

        args = (Fore.WHITE + six.text_type(self.source) + Style.RESET_ALL,
                Fore.WHITE + six.text_type(self.target) + Style.RESET_ALL,
                ign)
        return 'Replace `{}` → `{}`{}'.format(*args)

    def _run(self, marc_record):
        modified = 0

        for field in marc_record.search(self.source, self.ignore_extra_subfields):
            modified += field.replace(self.source, self.target)

        marc_record.remove_duplicates(self.target)

        return modified


@python_2_unicode_compatible
class InteractiveReplaceTask(SingleSourceConceptTask):
    """
    Replace a subject access or classification number field with another one
    (from a selection) in any given MARC record.

    Note: Exact matching only – will not replace fields having any additional
          subfields $b, $x, $y or $z. A search for "Fish" will not match the
          field "$a Fish $x Behaviour".
    """

    def __init__(self, source, targets, ignore_extra_subfields=False):
        super().__init__(source, ignore_extra_subfields)
        self.source.set_a_or_x_to('a')
        self.targets = deepcopy(targets)
        for target in self.targets:
            target.set_a_or_x_to('a')

    def _run(self, marc_record):
        utf8print()
        time.sleep(1)
        os.system('clear')
        if self.progress is not None:
            utf8print('{}[Record {:d} of {:d}]{}'.format(
                Fore.WHITE, self.progress['current'], self.progress['total'], Style.RESET_ALL
            ))
        utf8print('{}{} {}{}'.format(
            Fore.WHITE, marc_record.id, marc_record.title(), Style.RESET_ALL
        ))
        for field in marc_record.fields:
            if field.tag.startswith('6'):
                if field.sf('2') == self.source.sf['2']:
                    if field.match(self.source):
                        utf8print('  > {}{}{}'.format(Fore.YELLOW, field, Style.RESET_ALL))
                    else:
                        utf8print('    {}{}{}'.format(Fore.YELLOW, field, Style.RESET_ALL))
                else:
                    utf8print('    {}'.format(field))

        while True:
            targets = pick('Make a selection (or press Ctrl-C to abort)', self.targets, OrderedDict((
                ('REMOVE', 'None of them (remove the field)'),
            )))
            if 'REMOVE' in targets and len(targets) > 1:
                log.warning('Invalid selection. Please try again or press Ctrl-C to abort.')
            else:
                break

        if len(targets) == 0:
            log.info('Skipping this record')
            return 0

        tasks = []
        if 'REMOVE' in targets:
            tasks.append(DeleteTask([self.source], ignore_extra_subfields=self.ignore_extra_subfields))
        else:
            tasks.append(DeleteTask([self.source], ignore_extra_subfields=self.ignore_extra_subfields))
            for target in targets:
                tasks.append(AddTask(target))

        modified = 0
        for task in tasks:
            modified += task.run(marc_record)

        return modified

    def __str__(self):
        return 'Interactive replace'


@python_2_unicode_compatible
class ListTask(MultipleSourceConceptTask):
    """
    Do nothing except test if the MARC record contains the requested
    subject access or classification number field.

    Note: Exact matching only – will not replace fields having any additional
          subfields $b, $x, $y or $z. A search for "Fish" will not match the
          field "$a Fish $x Behaviour".
    """

    def __init__(self, concepts, show_titles=False, show_subjects=False, ignore_extra_subfields=True):
        super().__init__(concepts, ignore_extra_subfields)

        self.show_titles = show_titles
        self.show_subjects = show_subjects

    def __str__(self):
        return 'List titles having `{}`'.format(
            Fore.WHITE + '` and `'.join([six.text_type(x) for x in self.concepts]) + Style.RESET_ALL
        )

    def _run(self, marc_record):
        # if self.show_titles:
        #     utf8print('{}\t{}'.format(marc_record.id, marc_record.title()))
        # else:
        #     utf8print(marc_record.id)

        if self.show_subjects:
            for field in marc_record.fields:
                if field.tag.startswith('6'):
                    for concept in self.concepts:
                        if field.sf('2') == concept.sf['2']:
                            utf8print('  {}{}{}'.format(Fore.YELLOW, field, Style.RESET_ALL))
                        else:
                            utf8print('  {}{}{}'.format(Fore.CYAN, field, Style.RESET_ALL))

        return 0  # No, we didn't modify anything


@python_2_unicode_compatible
class DeleteTask(MultipleSourceConceptTask):
    """
    Delete one or more subject heading or classification number fields from any given MARC record.
    """

    def __init__(self, concepts, ignore_extra_subfields=False):
        super().__init__(concepts, ignore_extra_subfields)

    def __str__(self):
        return 'Delete ' + ' AND '.join([
            '`{}`'.format(Fore.WHITE + six.text_type(concept) + Style.RESET_ALL) for concept in self.concepts
        ])

    def _run(self, marc_record):
        removed = 0
        for concept in self.concepts:
            for field in marc_record.search(concept, self.ignore_extra_subfields):
                log.debug('Removing field: %s' % field)
                marc_record.remove_field(field)
                removed += 1

        return removed
        # Open question: should we also remove strings where sf['a'] is a component???


@python_2_unicode_compatible
class AddTask(Task):
    """
    Add a new subject access or classification number field to any given MARC record.
    """

    def __init__(self, target, match=False):
        self.target = deepcopy(target)
        self.target.set_a_or_x_to('a')

        # This task will either always match (if run alone)
        # or never match (if appending to another task)
        self._match = match

    def __str__(self):
        return 'Add `{}`'.format(Fore.WHITE + six.text_type(self.target) + Style.RESET_ALL)

    def match(self, marc_record):
        return self._match  # This task will only be run if some other task matches the record.

    def _run(self, marc_record):
        new_field = self.target.as_xml()

        idx = 0
        for field in marc_record.fields:
            try:
                node_tag = int(field.tag)
            except ValueError:  # Alma includes non-numeric tags like 'AVA'
                continue

            if node_tag > int(self.target.tag):
                break
            idx = marc_record.el.index(field.node)

        marc_record.el.insert(idx + 1, new_field)
        log.debug('Inserting field: %s' % self.target)

        marc_record.remove_duplicates(self.target)

        return 1



# coding=utf-8

from __future__ import unicode_literals

from .marc import Record
from .util import etree, parse_xml


class Bib(object):
    """ An Alma Bib record """

    def __init__(self, xml):
        self.orig_xml = xml
        self.init(xml)

    def init(self, xml):
        self.doc = parse_xml(xml)
        self.id = self.doc.findtext('mms_id')
        self.marc_record = Record(self.doc.find('record'))
        self.cz_link = self.doc.findtext('linked_record_id[@type="CZ"]') or None

    def xml(self):
        return (
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n%s' %
            etree.tounicode(self.doc)
        )

    def dump(self, filename):
        # Dump record to file
        with open(filename, 'wb') as file:
            file.write(etree.tostring(self.doc, pretty_print=True))



# coding=utf-8
from __future__ import unicode_literals
import difflib
import sys
from collections import OrderedDict
import vkbeautify
from colorama import Fore
from lxml import etree
from six import text_type
import questionary
import logging
import re
import pkg_resources  # part of setuptools
__version__ = pkg_resources.require('almar')[0].version

ANY_VALUE = '{ANY_VALUE}'

INTERACTIVITY_NONE = 0
INTERACTIVITY_STANDARD = 1
INTERACTIVITY_INCREASED = 2


def utf8print(txt=None):
    if txt is None:
        sys.stdout.write('\n')
    else:
        if sys.version_info < (3, 0):
            sys.stdout.write(('%s\n' % txt).encode('utf-8'))
        else:
            sys.stdout.write('%s\n' % txt)


def pick(msg, options, alpha_options=None):
    choices = OrderedDict()
    for i, option in enumerate(options):
        choice_text = text_type(option)
        choices[choice_text] = option
    if alpha_options is not None:
        for k, v in alpha_options.items():
            choices[v] = k

    answer = questionary.checkbox(msg, choices=[
        {'name': x} for x in choices.keys()
    ]).ask()

    return [choices[x] for x in answer]


def pick_one(msg, options, alpha_options=None):
    choices = OrderedDict()
    for i, option in enumerate(options):
        choice_text = text_type(option)
        choices[choice_text] = option
    if alpha_options is not None:
        for k, v in alpha_options.items():
            choices[v] = k

    answer = questionary.select(msg, choices.keys()).ask()

    return choices[answer]


def parse_xml(txt):
    if isinstance(txt, text_type):
        return etree.fromstring(txt.encode('utf-8'))
    return etree.fromstring(txt)


def normalize_term(term):
    # Normalize term so it starts with a capital letter. If the term is a subject string
    # fused by " : ", normalize all components.

    if term is None or len(term) == 0:
        return term

    return ' : '.join([component[0].upper() + component[1:] for component in term.strip().split(' : ')])


def term_match(term1, term2):
    return term1 == ANY_VALUE or term2 == ANY_VALUE or normalize_term(term1) == normalize_term(term2)


def color_diff(diff):
    for line in diff:
        if line.startswith('+'):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith('-'):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith('^'):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


def line_marc(root):

    st = []
    for node in root.xpath('//datafield'):
        t = '%s %s%s' % (node.get('tag'), node.get('ind1').replace(' ', '#'), node.get('ind2').replace(' ', '#'))
        for sf in node.findall('subfield'):
            t += ' $%s %s' % (sf.get('code'), sf.text)
        t += '\n'
        st.append(t)

    return st


def get_diff(src, dst):
    src = sorted(line_marc(etree.fromstring(src.encode('utf-8'))))
    dst = sorted(line_marc(etree.fromstring(dst.encode('utf-8'))))

    # src = vkbeautify.xml(src).splitlines(True)
    # dst = vkbeautify.xml(dst).splitlines(True)

    # returns list of unicode strings
    return list(difflib.unified_diff(src, dst, fromfile='Original', tofile='Modified'))


def format_diff(lines):
    return ''.join(color_diff(lines))


class ColorStripFormatter(logging.Formatter):

    def format(self, record):
        s = super(ColorStripFormatter, self).format(record)
        s = re.sub(r'\x1b\[[0-9;]*m', '', s)

        return s


class JobNameFilter(logging.Filter):

    jobname = ''

    def filter(self, record):
        record.jobname = self.jobname
        return True



# coding=utf-8
from __future__ import unicode_literals
import logging
from collections import OrderedDict
from copy import deepcopy
from six import python_2_unicode_compatible
from .util import etree, ANY_VALUE
log = logging.getLogger(__name__)


@python_2_unicode_compatible
class Concept(object):
    def __init__(self, tag, sf, ind1=None, ind2=None):
        if tag is None:
            raise ValueError('No tag given')
        if 'a_or_x' in sf and 'a' in sf:
            raise ValueError('Both a_or_x and a given')
        if 'a_or_x' in sf and 'x' in sf:
            raise ValueError('Both a_or_x and x given')
        self.tag = tag
        self.sf = deepcopy(sf)
        self.ind1 = ind1 or '?'
        self.ind2 = ind2 or '?'

        if self.sf.get('2') is None:
            raise RuntimeError('No vocabulary given')

    def __deepcopy__(self, memodict):
        return Concept(tag=self.tag, sf=self.sf, ind1=self.ind1, ind2=self.ind2)

    def has_subfield(self, code):
        if code in self.sf:
            return True
        if code in ['a', 'x'] and 'a_or_x' in self.sf:
            return True
        if code == 'a_or_x' and ('a' in self.sf or 'x' in self.sf):
            return True
        return False

    @property
    def components(self):
        return [
            value for key, value in self.sf.items()
            if key in ['a_or_x', 'a', 'b', 'x', 'y', 'z'] and value is not None
        ]

    @property
    def term(self):
        return ' : '.join([x for x in self.components if x != ANY_VALUE])

    def __str__(self):
        return self.tag + ' ' + ' '.join([
            '${} {}'.format(key[0], val)
            for key, val in self.sf.items() if val is not None
        ])

    def field(self):
        return {'tag': self.tag, 'sf': self.sf}

    def set_a_or_x_to(self, code):
        def get_key(key):
            if key == 'a_or_x':
                return code
            return key
        self.sf = OrderedDict([
            (get_key(key), val)
            for key, val in self.sf.items()
        ])

    @staticmethod
    def get_default_ind1(tag):
        # TODO: Vary on tag
        return ' '

    @staticmethod
    def get_default_ind2(tag):
        # TODO: Vary on tag
        return '7'

    def as_xml(self):

        if self.ind1 is not None and self.ind1 != '?':
            ind1 = self.ind1
        else:
            ind1 = self.get_default_ind1(self.tag)

        if self.ind2 is not None and self.ind2 != '?':
            ind2 = self.ind2
        else:
            ind2 = self.get_default_ind2(self.tag)

        new_field = etree.Element('datafield', {
            'tag': self.tag,
            'ind1': ind1,
            'ind2': ind2
        })
        for code, value in self.sf.items():
            subel = etree.SubElement(new_field, 'subfield', {'code': code})
            subel.text = value

        return new_field



# coding=utf-8
from __future__ import unicode_literals
import requests
import logging
import json
from colorama import Fore, Style
from .util import ANY_VALUE, pick, pick_one

log = logging.getLogger(__name__)


class Authorities(object):

    def __init__(self, vocabularies):
        self.vocabularies = vocabularies

    def authorize_concept(self, concept):
        if '2' not in concept.sf:
            raise ValueError('No vocabulary code (2) given!')
        if concept.sf['2'] in self.vocabularies:
            vocab = self.vocabularies[concept.sf['2']]
        else:
            log.info(Fore.RED + '✘' + Style.RESET_ALL + ' Could not authorize: %s', concept)
            return

        response = vocab.authorize_term(concept.term, concept.tag)

        if response.get('id') is not None:
            identifier = response.get('id')
            if concept.sf.get('0'):
                if concept.sf.get('0') == ANY_VALUE:
                    pass  # ignore ANY_VALUE
                elif identifier != concept.sf['0']:
                    identifier = pick_one(
                        'The $$0 value does not match the authority record id. Please select which to use',
                        [concept.sf['0'], identifier]
                    )
            concept.sf['0'] = identifier
            log.info(Fore.GREEN + '✔' + Style.RESET_ALL + ' Authorized: %s', concept)
        else:
            log.info(Fore.RED + '✘' + Style.RESET_ALL + ' Could not authorize: %s', concept)


class Vocabulary(object):

    marc_code = ''
    skosmos_code = ''

    def __init__(self, marc_code, id_service_url=None):
        self.marc_code = marc_code
        self.id_service_url = id_service_url

    def authorize_term(self, term, tag):
        # Lookup term with some id service to get the identifier to use in $0

        if term == '':
            return {}

        url = self.id_service_url.format(vocabulary=self.marc_code, term=term, tag=tag)
        response = requests.get(url)
        log.debug('Authority service response: %s', response.text)
        if response.status_code != 200 or response.text == '':
            return {}

        try:
            response = json.loads(response.text)
        except ValueError:
            log.error('ID lookup service returned: %s', response.text)
            return {}

        if 'error' in response and response.get('uri') != 'info:srw/diagnostic/1/61':
            log.warning('ID lookup service returned: %s', response['error'])

        return response



# coding=utf-8
from __future__ import unicode_literals

import logging
from copy import deepcopy
from datetime import datetime
import re

from colorama import Fore, Back, Style
from prompter import yesno
from tqdm import tqdm

from .sru import TooManyResults
from .task import AddTask, ReplaceTask, InteractiveReplaceTask, ListTask, DeleteTask, utf8print
from .util import INTERACTIVITY_NONE, INTERACTIVITY_STANDARD, INTERACTIVITY_INCREASED

log = logging.getLogger(__name__)
formatter = logging.Formatter('[%(asctime)s %(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%I:%S')


class Job(object):
    def __init__(self, action, source_concepts=[], target_concepts=[], sru=None, ils=None,
                 list_options=None, authorities=None, cql_query=None, grep=None):

        self.dry_run = False
        self.interactivity = INTERACTIVITY_STANDARD
        self.show_progress = True
        self.show_diffs = False
        self.list_options = list_options or {}

        self.records_changed = 0
        self.changes_made = 0

        self.sru = sru
        self.ils = ils
        self.authorities = authorities

        self.action = action
        self.source_concepts = source_concepts
        self.target_concepts = target_concepts

        self.job_name = datetime.now().isoformat()

        if (
            len(self.source_concepts) > 0 and
            self.source_concepts[0].tag == '648' and
            self.source_concepts[0].sf.get('2') == 'noubomn'
        ):
            raise RuntimeError('Editing 648 for noubomn is disabled until we get rid of the 650 duplicates')
            # log.info('Note: For the 648 field, we will also fix the 650 duplicate')

        self.authorize()
        for source_concept in source_concepts:
            log.debug('Source concept: %s', source_concept)
        for target_concept in target_concepts:
            log.debug('Target concept: %s', target_concept)

        def prepare_cql_query(source_concepts):
            query_parts = set()
            for concept in source_concepts:
                term = re.sub('[-–]', ' ', concept.term)  # replace hyphens and dashes with spaces
                query_parts.add('alma.subjects="%s"' % term)
                query_parts.add('alma.authority_vocabulary="%s"' % concept.sf['2'])

            query_parts = sorted(list(query_parts))
            return ' AND '.join(query_parts)

        self.cql_query = cql_query or prepare_cql_query(self.source_concepts)
        if self.cql_query == '':
            raise RuntimeError('No query given.')

        self.grep = grep
        if self.grep is not None:
            self.grep = self.grep.lower()

        self.steps = []
        self.generate_steps()

    @staticmethod
    def generate_replace_tasks(src, dst):
        """
        :type src: Concept
        :type dst: Concept
        """
        if len(src.components) == 1 and len(dst.components) == 1:
            if 'a_or_x' in src.sf and 'a_or_x' in dst.sf:
                tasks = []
                for code in ['a', 'x']:
                    src_copy = deepcopy(src)
                    dst_copy = deepcopy(dst)
                    src_copy.set_a_or_x_to(code)
                    dst_copy.set_a_or_x_to(code)
                    if code == 'a':
                        tasks.append(ReplaceTask(src_copy, dst_copy, False))  # exact match
                    tasks.append(ReplaceTask(src_copy, dst_copy, True))   # ignore extra subfields

                return tasks

        return [
            ReplaceTask(src, dst, False)
        ]

    def generate_steps(self):

        if self.action == 'custom':
            self.steps.append(DeleteTask(self.source_concepts))
            for target_concept in self.target_concepts:
                if len(self.source_concepts) == 0:
                    self.steps.append(AddTask(target_concept, match=True))
                else:
                    self.steps.append(AddTask(target_concept))

        elif self.action == 'add':
            for target_concept in self.target_concepts:
                self.steps.append(AddTask(target_concept, match=True))

        elif self.action == 'remove':
            # Delete
            self.steps.append(DeleteTask(self.source_concepts))

        elif self.action == 'interactive':
            self.steps.append(InteractiveReplaceTask(self.source_concepts[0], self.target_concepts))

        elif self.action == 'list':
            self.steps.append(ListTask(self.source_concepts, **self.list_options))

        elif self.action == 'replace':

            # Rename source concept to first target concept
            for step in self.generate_replace_tasks(self.source_concepts[0],
                                                    self.target_concepts[0]):
                self.steps.append(step)

            # Add remaining target concepts
            for target_concept in self.target_concepts[1:]:
                self.steps.append(AddTask(target_concept))

    def update_record(self, record, progress):
        """
        Update the record and save it back to Alma if any changes were made.
        Returns the number of changes made.
        """
        changes = 0
        for step in self.steps:
            changes += step.run(record.marc_record, progress)

        if changes == 0:
            return 0

        if self.interactivity == INTERACTIVITY_INCREASED and not yesno('Update this record?', default='yes'):
            return 0

        self.ils.put_record(record, interactive=self.interactivity != INTERACTIVITY_NONE, show_diff=self.show_diffs)

        return changes

    def authorize(self):
        if self.action in ['remove']:
            return

        # self.source_concept.authorize()
        if len(self.target_concepts) == 0:
            return
        self.authorities.authorize_concept(self.target_concepts[0])

        if '0' not in self.target_concepts[0].sf:
            log.warning('The (first) target term could not be authorized.')

        for target_concept in self.target_concepts[1:]:
            self.authorities.authorize_concept(target_concept)

    def start(self):

        if self.ils.name is not None:
            log.debug('Alma environment: %s', self.ils.name)

        log.debug('Planned steps:')
        for i, step in enumerate(self.steps):
            log.debug(' %d. %s' % ((i + 1), step))

        # ------------------------------------------------------------------------------------
        # Del 1: Søk mot SRU for å finne over alle bibliografiske poster med emneordet.
        # Vi må filtrere resultatlista i etterkant fordi
        #  - vi mangler en egen indeks for Realfagstermer, så vi må søke mot `alma.subjects`
        #  - søket er ikke presist, så f.eks. "Monstre" vil gi treff i "Mønstre"
        #
        # I fremtiden, når vi får $0 på alle poster, kan vi bruke indeksen `alma.authority_id`
        # i stedet.

        valid_records = set()
        pbar = None

        try:
            for marc_record in self.sru.search(self.cql_query):
                if pbar is None and self.show_progress and self.sru.num_records > 50:
                    pbar = tqdm(total=self.sru.num_records, desc='Filtering SRU results')

                log.debug('Checking record %s', marc_record.id)
                record_matching = False
                grep_matching = False
                for n, step in enumerate(self.steps):
                    step_matching = step.match(marc_record)

                    for field in marc_record.fields:
                        if self.grep is None or self.grep in str(field).lower():
                            grep_matching = True

                    if step_matching:
                        log.debug('Step %d did match', n)
                        record_matching = True
                    else:
                        log.debug('Step %d did not match', n)

                if record_matching and grep_matching:
                    valid_records.add(marc_record.id)

                if pbar is not None:
                    pbar.update()
            if pbar is not None:
                pbar.close()

        except TooManyResults:
            log.error((
                'More than 10,000 results would have to be checked, but the Alma SRU service does '
                'not allow us to retrieve more than 10,000 results. Annoying? Go vote for this:\n'
                'http://ideas.exlibrisgroup.com/forums/308173-alma/suggestions/'
                '18737083-sru-srw-increase-the-10-000-record-retrieval-limi'
            ))
            return []

        if len(valid_records) == 0:
            log.info('No matching catalog records found')
            return []
        elif self.action in ['interactive', 'list']:
            log.info('%d catalog records found', len(valid_records))
        else:
            log.info('%d catalog records to be changed', len(valid_records))

            if self.dry_run:
                log.warning('DRY RUN: No catalog records will actually be changed!')

            if not self.dry_run and self.interactivity == INTERACTIVITY_STANDARD and not yesno('Continue?', default='yes'):
                log.info('Job aborted')
                return []

        # ------------------------------------------------------------------------------------
        # Del 2: Nå har vi en liste over MMS-IDer for bibliografiske poster vi vil endre.
        # Vi går gjennom dem én for én, henter ut posten med Bib-apiet, endrer og poster tilbake.

        self.records_changed = 0
        self.changes_made = 0
        for idx, mms_id in enumerate(valid_records):
            if self.action not in ['list', 'interactive']:
                log.info('Record %d/%d: %s', idx + 1, len(valid_records), mms_id)

            record = self.ils.get_record(mms_id)

            if self.list_options.get('show_titles'):
                utf8print('{}\t{}'.format(record.marc_record.id, record.marc_record.title()))

            if self.list_options.get('show_subjects'):
                for field in record.marc_record.fields:
                    if field.tag.startswith('6'):
                        if len(self.source_concepts) > 0 and field.sf('2') == self.source_concepts[0].sf['2']:
                            utf8print('  {}{}{}'.format(Fore.YELLOW, field, Style.RESET_ALL))
                        else:
                            utf8print('  {}{}{}'.format(Fore.CYAN, field, Style.RESET_ALL))

            c = self.update_record(record, progress={'current': idx + 1, 'total': len(valid_records)})

            if c > 0:
                self.records_changed += 1
                self.changes_made += c

        return valid_records



# coding=utf-8
from __future__ import unicode_literals
import logging
from io import BytesIO
from prompter import yesno
from requests import Session, HTTPError
from textwrap import dedent

from .util import get_diff, format_diff
from .bib import Bib

log = logging.getLogger(__name__)


class LibrarySystem(object):

    def get_record(self, record_id):
        raise NotImplementedError()

    def put_record(self, record):
        raise NotImplementedError()


class Alma(LibrarySystem):

    name = None

    def __init__(self, api_region, api_key, name=None, dry_run=False):
        self.api_region = api_region
        self.api_key = api_key
        self.name = name
        self.dry_run = dry_run
        self.session = Session()
        self.session.headers.update({'Authorization': 'apikey %s' % api_key})
        self.base_url = 'https://api-{region}.hosted.exlibrisgroup.com/almaws/v1'.format(region=self.api_region)

    def url(self, path, **kwargs):
        return self.base_url.rstrip('/') + '/' + path.lstrip('/').format(**kwargs)

    def get_record(self, record_id):
        """
        Get a Bib record from Alma

        :type record_id: string
        """
        response = self.session.get(self.url('/bibs/{mms_id}', mms_id=record_id))
        response.raise_for_status()
        record = Bib(response.text)
        if record.id != record_id:
            raise RuntimeError('Response does not contain the requested MMS ID. %s != %s'
                               % (record.id, record_id))
        return record

    def put_record(self, record, interactive=True, show_diff=False):
        """
        Store a Bib record to Alma

        :param show_diff: bool
        :param interactive: bool
        :type record: Bib
        """
        if record.cz_link is not None:
            log.warning(dedent(
                '''\
                Encountered a Community Zone record. Updating such records through the API will
                currently cause them to be de-linked from CZ, which is probably not what you want.
                Until Ex Libris fixes this, you're best off editing the record manually in Alma.\
                '''))

            if not interactive or yesno('Do you want to update the record and break CZ linkage?', default='no'):
                log.warning(' -> Skipping this record. You should update it manually in Alma!')
                return

            log.warning(' -> Updating the record. The CZ connection will be lost!')

        post_data = record.xml()
        diff = get_diff(record.orig_xml, post_data)
        additions = len([x for x in diff[2:] if x[0] == '+'])
        deletions = len([x for x in diff[2:] if x[0] == '-'])
        if show_diff:
            log.info('%d line(s) removed, %d line(s) added:\n%s', deletions, additions, format_diff(diff))
        else:
            log.debug('%d line(s) removed, %d line(s) added:\n%s', deletions, additions, format_diff(diff))

        if not self.dry_run:
            try:
                response = self.session.put(self.url('/bibs/{mms_id}', mms_id=record.id),
                                            data=BytesIO(post_data.encode('utf-8')),
                                            headers={'Content-Type': 'application/xml'})
                response.raise_for_status()
                record.init(response.text)

            except HTTPError:
                msg = '*** Failed to save record %s --- Please try to edit the record manually in Alma ***'
                log.error(msg, record.id)



# coding=utf-8
from __future__ import unicode_literals

import argparse
import getpass
from collections import OrderedDict

import colorama
import time
from coloredlogs import ColoredFormatter
import logging
import logging.config
import re
import os
import sys
from io import open  # pylint: disable=redefined-builtin
from hashlib import sha1

import yaml
from six import text_type
from raven import Client
from six import binary_type

from . import __version__
from .authorities import Vocabulary, Authorities
from .alma import Alma
from .concept import Concept
from .job import Job
from .sru import SruClient
from .util import ANY_VALUE, INTERACTIVITY_NONE, INTERACTIVITY_STANDARD, INTERACTIVITY_INCREASED
from .util import ColorStripFormatter, JobNameFilter

raven_client = None


def configure_logging(config, jobname, verbose=False):
    use_colors = sys.stdout.isatty()
    level = logging.DEBUG if verbose else logging.INFO
    formatter_options = {
        'fmt': '%(asctime)s %(levelname)-8s %(message)s',
        'datefmt': '%Y-%m-%d %H:%I:%S',
    }

    if use_colors:
        colorama.init(autoreset=True)
    else:
        # We're being piped, so skip colors
        colorama.init(strip=True)

    logging.config.dictConfig(config)

    # Get root logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # Add stream handler and formatter
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter_type = ColoredFormatter if use_colors else ColorStripFormatter
    handler.setFormatter(formatter_type(**formatter_options))

    # Configure JobNameFilter
    JobNameFilter.jobname = jobname[:10]

    logger.addHandler(handler)

    # Increase logging level for dependencies
    logging.getLogger('requests').setLevel(logging.WARNING)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)


def ensure_unicode(arg):
    if isinstance(arg, binary_type):
        return arg.decode('utf-8')
    return arg


def parse_args(args, default_env=None):
    parser = argparse.ArgumentParser(prog='almar',
                                     description='Edit or remove subject fields in Alma catalog records. '
                                     'By default, almar will ask you once to confirm before making any edits. '
                                     'This can be changed by using the -n or -i flags, see below.')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    parser.add_argument('-e', '--env', dest='env', nargs='?',
                        help='Environment from config file. Default: {}'.format(default_env or '(none)'),
                        default=default_env)

    parser.add_argument('-d', '--dry-run', '--dry_run', dest='dry_run', action='store_true',
                        help='Dry run without doing any edits.')

    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Show more output')

    parser.add_argument('-n', '--non-interactive', '--non_interactive', dest='non_interactive', action='store_true',
                        help='Non-interactive mode: Never ask for confirmation, always use defaults.')

    parser.add_argument('-i', '--interactive', dest='interactive', action='store_true',
                        help='Interactive mode: ask to confirm each change.')

    parser.add_argument('--diffs', dest='show_diffs', action='store_true',
                        help='Show diffs (deprecated option, now enabled by default).')

    parser.add_argument('--cql', dest='cql_query', nargs='?',
                        help=('Custom CQL query to specify which records to be checked. '
                              'Example: --cql \'alma.all_for_ui = "some identifier"\'')
                        )

    parser.add_argument('--titles', dest='show_titles', action='store_true',
                        help='Show titles (deprecated option, now enabled by default)')
    parser.add_argument('--subjects', dest='show_subjects', action='store_true', help='Show subject fields')

    parser.add_argument('--grep', dest='grep', nargs='?',
                        help=('Filter the result list by some string.'
                              'Example: --grep \'some text\'')
                        )

    parser.add_argument('--rem', dest='remove', action='append', default=[], help='Term to remove (can be repeated).')
    parser.add_argument('--add', dest='add', action='append', default=[], help='Term to add (can be repeated).')

    subparsers = parser.add_subparsers(title='subcommands')

    # Create parser for the "replace" command
    parser_move = subparsers.add_parser('replace', help='Replace/rename/move subject field')
    # TODO: , aliases=['rename', 'move'] added in Python 3.5

    parser_move.add_argument('term', nargs=1, help='Term to search for')
    parser_move.add_argument('new_terms', nargs='+', default='', help='Replacement terms')
    parser_move.set_defaults(action='replace')

    # Create parser for the "remove" command
    parser_del = subparsers.add_parser('remove', help='Remove subject field')
    # TODO: # , aliases=['delete'] added in Python 3.5

    parser_del.add_argument('term', nargs=1, help='Term to remove')
    parser_del.set_defaults(action='remove')

    # Create parser for the "add" command
    parser_add = subparsers.add_parser('add', help='Add subject fields')
    parser_add.add_argument('new_terms', nargs='+', help='Terms to add')
    parser_add.set_defaults(action='add')

    # Create parser for the "interactive" command
    parser_int = subparsers.add_parser('interactive', help='Interactive reclassification')
    parser_int.add_argument('term', nargs=1, help='Term to search for')
    parser_int.add_argument('new_terms', nargs='+', default='', help='Replacement terms')
    parser_int.set_defaults(action='interactive')

    # Create parser for the "list" command
    parser_list = subparsers.add_parser('list', help='List documents')
    parser_list.add_argument('term', nargs=1, help='Term to search for')
    parser_list.set_defaults(action='list')

    # Parse
    args = parser.parse_args(args)

    # Deprecated options
    args.show_diffs = True
    args.show_titles = True

    if 'new_terms' not in args:
        args.new_terms = []

    if 'action' not in args:
        if len(args.remove) == 0 and len(args.add) == 0:
            parser.error('Please specify an action or one or more --rem or --add clauses.')
        args.action = 'custom'
    else:
        if len(args.remove) != 0 or len(args.add) != 0:
            parser.error('--rem or --add cannot be used when an action is specified.')

    if args.interactive and args.non_interactive:
        parser.error('-n and -i are mutually exclusive')

    if args.env is not None:
        args.env = args.env.strip()

    args.env = ensure_unicode(args.env)

    args.terms = [ensure_unicode(x) for x in args.remove]
    if 'term' in args:
        args.terms.append(ensure_unicode(args.term[0]))

    args.new_terms = [ensure_unicode(x) for x in args.add] + [ensure_unicode(x) for x in args.new_terms]

    return args


def normalize_ind(value):
    if value == '#':
        return ' '
    return value


def parse_advanced_input(value):
    log = logging.getLogger()
    m = re.match(r'^(?P<tag>[0-9]{3}) (?P<ind1>[0-9#])(?P<ind2>[0-9#]) (?P<sf>\$\$.*)$', value)
    if not m:
        log.error('Invalid input format')
        sys.exit(1)
    sf = OrderedDict()
    for m2 in re.finditer(r'(?P<code>[a-z0-9]) (?P<val>[^\$]+)', m.group('sf')):
        sf[m2.group('code')] = m2.group('val').strip()

    if len(sf) == 0:
        log.error('Invalid input format')
        sys.exit(1)

    return {
        'tag': m.group('tag'),
        'ind1': normalize_ind(m.group('ind1')),
        'ind2': normalize_ind(m.group('ind2')),
        'sf': sf,
    }


def parse_components(streng):
    components = streng.split(' : ')
    sf = OrderedDict()
    if len(components) == 1:
        sf['a_or_x'] = components[0]
    elif len(components) == 2:
        sf['a'] = components[0]
        sf['x'] = components[1]
    if len(components) > 2:
        raise RuntimeError('Strings with more than two components are not supported')
    return sf


def get_concept(term, default_vocabulary, default_tag=None, default_term=None):

    default_tag = default_tag or '650'

    # 1) Advanced syntax
    if '$$' in term:
        return Concept(**parse_advanced_input(term))

    # 2) Just tag
    match = re.match('^([0-9]{3})$', term)
    if match:
        if default_term is None:
            raise RuntimeError('No source term specified')
        sf = parse_components(default_term)
        sf['2'] = default_vocabulary
        return Concept(match.group(1), sf)

    # 3) Tag and term
    match = re.match('^([0-9]{3}) (.+)$', term)
    if match:
        sf = parse_components(match.group(2))
        sf['2'] = default_vocabulary
        return Concept(match.group(1), sf)

    # 4) Just term
    sf = parse_components(term)
    sf['2'] = default_vocabulary
    return Concept(default_tag, sf)


def job_args(config=None, args=None):

    vocabularies = {}
    for vocab in config.get('vocabularies', []):
        vocabularies[ensure_unicode(vocab['marc_code'])] = Vocabulary(
            ensure_unicode(vocab['marc_code']),
            ensure_unicode(vocab.get('id_service')),
        )
    default_vocabulary = ensure_unicode(config['default_vocabulary'])

    source_concepts = [
        get_concept(term, default_vocabulary)
        for term in args.terms
    ]

    target_concepts = []
    if args.action == 'replace':
        target_concepts.append(get_concept(args.new_terms[0], default_vocabulary,
                                           default_term=source_concepts[0].term,
                                           default_tag=source_concepts[0].tag))

        for new_term in args.new_terms[1:]:
            target_concepts.append(get_concept(new_term, default_vocabulary,
                                               default_tag=source_concepts[0].tag))

    elif args.action in ['interactive', 'custom']:
        target_concepts = [
            get_concept(term, default_vocabulary)
            for term in args.new_terms
        ]

    list_options = {}

    """ Caveat 1:

    We will do fuzzy matching (matching either $a or $x) only if both the source
    and target supports it
    """
    log = logging.getLogger()

    if len(source_concepts) == 1 and len(target_concepts) > 0:
        if 'a_or_x' in source_concepts[0].sf and 'a_or_x' not in target_concepts[0].sf:
            source_concepts[0].set_a_or_x_to('a')

        if 'a_or_x' in target_concepts[0].sf and 'a_or_x' not in source_concepts[0].sf:
            target_concepts[0].set_a_or_x_to('a')

    """ Caveat 2:

    If more than one source or target concept is involved, don't do fuzzy matching.
    """
    if len(source_concepts) > 1 or len(target_concepts) > 1:
        for source_concept in source_concepts:
            source_concept.set_a_or_x_to('a')
        for target_concept in target_concepts:
            target_concept.set_a_or_x_to('a')

    """ Caveat 3:

    If a tag move is involved, don't do fuzzy matching
    """
    if len(source_concepts) == 1 and len(target_concepts) > 0:
        if source_concepts[0].tag != target_concepts[0].tag:
            if 'a_or_x' in source_concepts[0].sf:
                source_concepts[0].set_a_or_x_to('a')
            if 'a_or_x' in target_concepts[0].sf:
                target_concepts[0].set_a_or_x_to('a')

    """ Caveat 4a:

    If a subfield exists in the source query, but not in the target query,
    we interpret that as a request for removing the subfield.
    As an example, the command

        almar replace '650 $$a TermA $$b TermB' '650 $$a TermC'

    should replace 'TermA' with 'TermC' in $$a and remove $$b.

    Developer note: This should be run before caveat 4 below, since we don't
    want to remove identifiers! (This is covered by tests)
    """
    if len(source_concepts) == 1:
        for target_concept in target_concepts:
            # loop over all target concepts because of InteractiveReplaceTask
            for code in source_concepts[0].sf:
                if not target_concept.has_subfield(code) and code != '0':
                    log.debug('Adding explicit "%s: None" to target concept %s', code, target_concept)
                    target_concept.sf[code] = None  # meaning NO_VALUE

    """ Caveat 4b:

    If a subfield (except for $0) exists in the target query, but not in the
    source query, we should not match fields already having some value for
    that subfield. As an example, the command

        almar replace '650 $$a TermA' '650 $$a TermB $$b TermC'

    should not match fields having '650 $$a TermA $$b SomeValue'

    Note: If there are multiple targets with varying number of components,
    this still applies for any subfield found in *any* of the targets. E.g.

        almar replace '650 $$a TermA' '650 $$a TermB' '650 $$a TermB $$b TermC'

    would not match '650 $$a TermA $$b SomeValue' Could this be counterintuitive?

    Developer note: This should be run before caveat 4 below
    """
    if len(source_concepts) == 1:
        for target_concept in target_concepts:
            # loop over all target concepts because of InteractiveReplaceTask
            for code in target_concept.sf:
                if not source_concepts[0].has_subfield(code) and code != '0':
                    log.debug('Adding explicit "%s: None" to source concept %s', code, source_concepts[0])
                    source_concepts[0].sf[code] = None  # meaning NO_VALUE

    """ Caveat 5:

    Some fields will have $0 values, but many won't, so we cannot require
    a $0 value at this point. If you want to match only fields with a given
    $0 value, use the advanced syntax:

        almar replace '650 $$a Test $$b Test $$0 identifer' ...

    """
    for source_concept in source_concepts:
        if '0' not in source_concept.sf:
            source_concept.sf['0'] = ANY_VALUE

    list_options['show_titles'] = args.show_titles
    list_options['show_subjects'] = args.show_subjects

    return {
        'action': args.action,
        'source_concepts': source_concepts,
        'target_concepts': target_concepts,
        'list_options': list_options,
        'cql_query': args.cql_query,
        'grep': args.grep,
        'authorities': Authorities(vocabularies)
    }


def get_config_filename():
    possible_file_locations = ['./almar.yml', './lokar.yml', os.path.expanduser('~/.almar.yml')]

    for filename in possible_file_locations:
        if os.path.exists(filename):
            return filename


def get_config():
    log = logging.getLogger()

    filename = get_config_filename()
    if filename is None:
        log.error('Could not find "almar.yml" configuration file. See https://github.com/scriptotek/almar for help.')
        sys.exit(1)
    try:
        with open(filename) as fp:
            config = yaml.load(fp)
    except IOError:
        log.error('Could not read configuration file "%s"', filename)
        sys.exit(1)

    return config


def run(config, argv):
    global raven_client

    username = getpass.getuser()

    logging_defaults = {
        'version': 1,
        'disable_existing_loggers': False,
        'root': {
            'level': 'INFO',
        }
    }

    # Python 2/3: normalize to unicode strings
    argv = [x.decode('utf-8') if isinstance(x, binary_type) else x for x in argv]

    sha_input = u' '.join([config.get('default_env'), text_type(time.time())] + argv)
    jobname = sha1(sha_input.encode('utf-8')).hexdigest()

    # Note: configure_logging will add a StreamHandler for stdout
    args = parse_args(argv, config.get('default_env'))

    configure_logging(config.get('logging', logging_defaults), jobname, args.verbose)
    log = logging.getLogger()
    if sys.version_info < (3, 5):
        log.error('Sorry, Python < 3.5 is not supported.')
        sys.exit(1)
    log.debug('Starting job %s as %s', jobname, username)

    jargs = job_args(config, args)

    if config.get('sentry') is not None:
        raven_client = Client(config['sentry']['dsn'])
        raven_client.context.merge({'user': {
            'username': username
        }})
    try:
        def get_env(config, args):
            if args.env is None:
                log.error('No environment specified and no default environment found in configuration file')
                sys.exit(1)

            for env in config.get('env', []):
                if env['name'] == args.env:
                    return env

            log.error('Environment "%s" not found in configuration file', args.env)
            sys.exit(1)

        env = get_env(config, args)

        sru = SruClient(env['sru_url'], args.env)
        alma = Alma(env['api_region'], env['api_key'], args.env, dry_run=args.dry_run)

        job = Job(sru=sru, ils=alma, **jargs)
        job.dry_run = args.dry_run
        if args.non_interactive:
            job.interactivity = INTERACTIVITY_NONE
        elif args.interactive:
            job.interactivity = INTERACTIVITY_INCREASED
        else:
            job.interactivity = INTERACTIVITY_STANDARD

        job.verbose = args.verbose
        job.show_diffs = args.show_diffs

        concepts = jargs['source_concepts'] + jargs['target_concepts']
        jobdesc = '%s %s' % (jargs['action'], ' '.join(["'%s'" % text_type(x) for x in concepts]))

        log.debug('Job arguments: %s', jobdesc)

        job.start()

        if job.changes_made > 0:
            log.info('Job %s completed. Made %d changes to %d records', jobname, job.changes_made, job.records_changed)

            summary = logging.getLogger('summary')
            summary.info('%s - %s - %s - Made %d changes to %d records',
                         jobname, username, jobdesc, job.changes_made, job.records_changed)

    except Exception:  # # pylint: disable=broad-except
        if raven_client is not None:
            raven_client.captureException()
        log.exception('Uncaught exception:')


def main():
    run(get_config(), sys.argv[1:])


if __name__ == '__main__':
    main()



# coding=utf-8
from __future__ import unicode_literals

import warnings
import logging
from collections import OrderedDict
from copy import deepcopy

from six import python_2_unicode_compatible

from .util import term_match, parse_xml, ANY_VALUE

log = logging.getLogger(__name__)


def deprecated(func):
    '''This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.'''
    def new_func(*args, **kwargs):
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning)
        return func(*args, **kwargs)
    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)

    return new_func


@python_2_unicode_compatible
class Subfield(object):
    """ A Marc21 subfield """
    def __init__(self, node):
        self.node = node

    @property
    def code(self):
        return self.node.get('code')

    @property
    def text(self):
        return self.node.text

    @text.setter
    def text(self, value):
        self.node.text = value

    def __str__(self):
        return self.text

    # def __cmp__(self, other):
    #     if self.code != other.code:
    #         return False

    #     if self.value == ANY_VALUE or other.value == ANY_VALUE
    #         return True

    #     and self.value == other.value


@python_2_unicode_compatible
class Field(object):
    """ A Marc21 field """

    def __init__(self, node):
        self.node = node

    @property
    def tag(self):
        return self.node.get('tag')

    @property
    def ind1(self):
        return self.node.get('ind1')

    @property
    def ind2(self):
        return self.node.get('ind2')

    def __str__(self):
        items = [self.tag, self.ind1.replace(' ', '#') + self.ind2.replace(' ', '#')]
        for subfield in self.node:
            items.append('$%s %s' % (subfield.attrib['code'], subfield.text))
        return ' '.join(items)

    @property
    def subfields(self):
        return self.get_subfields()

    def get_subfields(self, source_code=None):
        for node in self.node.findall('subfield'):
            if source_code is None or source_code == node.get('code'):
                yield Subfield(node)

    def sf(self, code=None):
        # return text of first matching subfield or None
        for node in self.get_subfields(code):
            return node.text

    @deprecated  # use sf instead
    def subfield_text(self, code):
        return self.sf(code)

    def set_tag(self, value):
        if self.node.get('tag') != value:
            log.debug('CHANGE: Set tag to %s in `%s`', value, self)
            self.node.set('tag', value)
            return 1
        return 0

    def set_ind1(self, value):
        if value is not None and value != '?' and self.node.get('ind1') != value:
            log.debug('CHANGE: Set ind1 to %s in `%s`', value, self)
            self.node.set('ind1', value)
            return 1
        return 0

    def set_ind2(self, value):
        if value is not None and value != '?' and self.node.get('ind2') != value:
            log.debug('CHANGE: Set ind2 to %s in `%s`', value, self)
            self.node.set('ind2', value)
            return 1
        return 0

    def match(self, concept, ignore_extra_subfields=False):
        """
        Return True if 'field' matches this concept
        """

        if not self.tag.startswith(concept.tag):
            return False

        if concept.ind1 != '?' and self.ind1 != concept.ind1:
            return False

        if concept.ind2 != '?' and self.ind2 != concept.ind2:
            return False

        for code, sf_value in concept.sf.items():
            if sf_value != ANY_VALUE and not term_match(sf_value, self.sf(code)):
                return False

        if not ignore_extra_subfields:
            for subfield in self.subfields:
                if subfield.code not in concept.sf and subfield.code not in ['0', '9']:
                    return False

        return True

    def replace(self, source, target):
        """
        Replace field with target
        """

        modified = 0

        modified += self.set_tag(target.tag)
        modified += self.set_ind1(target.ind1)
        modified += self.set_ind2(target.ind2)
        modified += self.update_subfields(source, target)

        return modified

    def update_subfields(self, source, target):
        modified = 0
        idx = 0

        for code, target_value in target.sf.items():
            source_value = source.sf.get(code)
            found_subfield = False

            for subfield in self.get_subfields(code):
                if term_match(source_value, subfield.text):
                    found_subfield = True
                    idx = self.node.index(subfield.node) + 1

                    if target_value is None:  # NoValue
                        # Subfield should be removed
                        log.debug('CHANGE: Removing `$%s %s` from `%s`', code, subfield.text, self)
                        self.node.remove(subfield.node)
                        modified += 1

                    elif subfield.text != target_value:
                        # Subfield value should be updated
                        log.debug('CHANGE: Changing $%s from "%s" to "%s" in `%s`',
                                  code, subfield.text, target_value, self)
                        subfield.text = target_value
                        modified += 1

            if not found_subfield and target_value is not None:
                # Add subfield
                log.debug('CHANGE: Adding `$%s %s` to `%s`', code, target_value, self)
                self.node.insert(idx,
                                 parse_xml('<subfield code="%s">%s</subfield>' % (code, target_value))
                                 )
                modified += 1

        return modified


class Record(object):
    """ A Marc21 record """

    def __init__(self, el):
        # el: xml.etree.ElementTree.Element
        self.el = el

    @property
    def id(self):
        return self.el.findtext('./controlfield[@tag="001"]')

    @property
    def fields(self):
        return self.get_fields()

    def get_fields(self):
        for node in self.el.findall('datafield'):
            yield Field(node)

    def search(self, concept, ignore_extra_subfields=False):
        """
        Return fields matching the Concept
        """
        for field in self.fields:
            if field.match(concept, ignore_extra_subfields):
                yield field

    def remove_duplicates(self, concept, ignore_extra_subfields=False):
        concept = deepcopy(concept)

        concept.sf['0'] = ANY_VALUE

        dups = 0
        candidates = {}

        # query2 = {}
        # for key, value in query.items():
        #     if value is None:
        #         query2[key] = None
        #     elif six.text_type(value) == value:
        #         query2[key] = value
        #     elif 'replace' in value:
        #         query2[key] = value['replace']
        #     else:
        #         query2[key] = value['search']

        matches = list(self.search(concept, ignore_extra_subfields))
        if len(matches) > 1:
            # Sort fields with $0 first, since we prefer to keep those
            matches = sorted(matches, key=lambda x: x.sf('0') or '', reverse=True)
            for match in matches[1:]:
                log.info('Target subject already existed on the record, ignoring duplicate: %s', match)
                self.remove_field(match)
                dups += 1

        return dups

    def remove_field(self, field):
        # field: Field
        self.el.remove(field.node)

    def title(self):

        out = self.el.find('./datafield[@tag="245"]/subfield[@code="a"]').text

        node = self.el.find('./datafield[@tag="245"]/subfield[@code="b"]')
        if node is not None:
            out = out.rstrip(' :') + ' : ' + node.text

        node = self.el.find('./datafield[@tag="245"]/subfield[@code="p"]')
        if node is not None:
            out = out.rstrip(' /:.') + '. ' + node.text

        node = self.el.find('./datafield[@tag="245"]/subfield[@code="n"]')
        if node is not None:
            out = out.rstrip(' /:.') + '. ' + node.text

        node = self.el.find('./datafield[@tag="245"]/subfield[@code="c"]')
        if node is not None:
            out = out.rstrip(' /') + ' / ' + node.text

        out = out.rstrip('.') + '.'

        node = self.el.find('./datafield[@tag="264"]/subfield[@code="c"]')
        if node is None:
            node = self.el.find('./datafield[@tag="260"]/subfield[@code="c"]')
            if node is not None:
                out += ' ' + node.text

        return out



# coding=utf-8
from __future__ import unicode_literals

import logging

import requests

from .marc import Record
from .util import parse_xml

log = logging.getLogger(__name__)

NSMAP = {
    'e20': 'http://explain.z3950.org/dtd/2.0/',
    'e21': 'http://explain.z3950.org/dtd/2.1/',
    'srw': 'http://www.loc.gov/zing/srw/',
    'diag': 'http://www.loc.gov/zing/srw/diagnostic/',
}


class SruErrorResponse(RuntimeError):
    pass


class TooManyResults(RuntimeError):
    pass


class SruClient(object):

    def __init__(self, endpoint_url, name=None):
        self.endpoint_url = endpoint_url
        self.name = name
        self.record_no = 0  # from last response
        self.num_records = 0  # from last response

    def request(self, query, start_record):
        response = requests.get(self.endpoint_url, params={
            'version': '1.2',
            'operation': 'searchRetrieve',
            'startRecord': start_record,
            'maximumRecords': '50',
            'query': query,
        })
        return response.text

    def search(self, query):
        log.debug('SRU search: %s', query)
        # A searchRetrieve generator that yields MarcRecord objects
        start_record = 1
        while True:
            response = self.request(query, start_record)

            # Fix for the sudden addition of namespaces to the SRU response.
            # The problem is that the Bibs API still don't use namespaces,
            # so by removing the namespace the XML is compatible with the Bibs API.
            txt = response.replace('xmlns="http://www.loc.gov/MARC21/slim"', 'xmlns=""')

            root = parse_xml(txt)  # Takes ~ 4 seconds for 50 records!

            for diagnostic in root.findall('srw:diagnostics/diag:diagnostic', namespaces=NSMAP):
                raise SruErrorResponse(diagnostic.findtext('diag:message', namespaces=NSMAP))

            self.num_records = int(root.findtext('srw:numberOfRecords', namespaces=NSMAP))
            if self.num_records > 10000:
                raise TooManyResults()

            for record in root.iterfind('srw:records/srw:record', namespaces=NSMAP):
                self.record_no = int(record.findtext('srw:recordPosition', namespaces=NSMAP))

                yield Record(record.find('srw:recordData/record', namespaces=NSMAP))

            nrp = root.find('srw:nextRecordPosition', namespaces=NSMAP)
            if nrp is not None:
                start_record = nrp.text
            else:
                break  # Enden er nær, den er faktisk her!



#!/usr/bin/env python
# encoding=utf-8
import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setup(name='almar',
      version='0.8.0',  # Use bumpversio to update
      description='Search and replace for subject fields in Alma records.',
      long_description=README,
      long_description_content_type='text/markdown',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='marc alma',
      author='Dan Michael O. Heggø',
      author_email='d.m.heggo@ub.uio.no',
      url='https://github.com/scriptotek/almar',
      license='MIT',
      install_requires=['six',
                        'requests',
                        'tqdm',
                        'prompter',
                        'pyyaml',
                        'colorama',
                        'coloredlogs',
                        'raven',
                        'vkbeautify',
                        'lxml',
                        'questionary',
                        ],
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'pytest-pycodestyle', 'pytest-cov', 'responses', 'mock'],
      entry_points={'console_scripts': ['almar=almar.almar:main']},
      options={
          'build_scripts': {
              'executable': '/usr/bin/env python',
          },
      },
      packages=['almar']
      # data_files=[(AppDirs('Lokar').user_config_dir, ['almar.cfg'])]
      )
